#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <avr/pgmspace.h>

#include "interpreter.h"
#include "lcd.h"
#include "keypad.h"
#include "hex_font.h"

uint8_t interpreter_memory[INTERPRETER_MEMORY_MAX - INTERPRETER_MEMORY_MIN + 1] __attribute__ ((section (".noinit")));

uint8_t interpreter_registers[16] __attribute__ ((section (".noinit")));
uint8_t interpreter_hp48_flags[16] __attribute__ ((section (".noinit")));

uint16_t interpreter_stack[INTERPRETER_STACK_LEVELS] __attribute__ ((section (".noinit")));
uint8_t interpreter_stack_pointer __attribute__ ((section (".noinit")));

uint16_t interpreter_program_counter __attribute__ ((section (".noinit")));
uint16_t interpreter_address_register __attribute__ ((section (".noinit")));

uint8_t interpreter_flags __attribute__ ((section (".noinit")));

// General-purpose delay timer. It decrements 60 times per second until it reaches zero.
uint8_t interpreter_delay_timer __attribute__ ((section (".noinit")));

// Delay timer that makes a sound when non-zero. It decrements 60 times per second until it reaches zero.
uint8_t interpreter_sound_timer __attribute__ ((section (".noinit")));

// Resets the interpreter by unloading the program from memory and clearing registers to default values.
void interpreter_reset(void) {
	interpreter_program_counter = 0x0200;
	interpreter_address_register = 0x0000;
	memset(interpreter_memory, 0, INTERPRETER_MEMORY_MAX - INTERPRETER_MEMORY_MIN + 1);
	memset(interpreter_registers, 0, 16);
	memset(interpreter_hp48_flags, 0, 16);
	memcpy_P(interpreter_memory, interpreter_hex_font_data, INTERPRETER_FONT_DATA_LENGTH);
	interpreter_stack_pointer = 0;
	interpreter_flags = INTERPRETER_FLAGS_DEFAULT;
	interpreter_delay_timer = 0;
	interpreter_sound_timer = 0;
	lcd_clear();
}

// Loads a program into the interpreter from ROM.
void interpreter_load_P(const uint8_t* program, size_t length) {
	memcpy_P(interpreter_memory + (0x200 - INTERPRETER_MEMORY_MIN), program, length);
}

// Fetches and executes a single CHIP-8 instruction.
void interpreter_execute_instruction(void) {

	// Fetch the instruction.
	
	//TODO: program counter range check
	uint16_t current_instruction = (interpreter_memory[interpreter_program_counter - INTERPRETER_MEMORY_MIN] << 8) | interpreter_memory[interpreter_program_counter + 1 - INTERPRETER_MEMORY_MIN];
	interpreter_program_counter += 2;

	switch (current_instruction & 0xF000) {
		case 0x0000:
			switch (current_instruction & 0xFF) {
				case 0xE0:
					lcd_clear();
					break;
				case 0xFE:
					interpreter_flags &= (uint8_t)~INTERPRETER_FLAGS_SCHIP;
					break;
				case 0xFF:
					interpreter_flags |= INTERPRETER_FLAGS_SCHIP;
					break;
				case 0xEE:
					// TODO: stack underflow check.
					interpreter_program_counter = interpreter_stack[--interpreter_stack_pointer];
					break;
				case 0xFC: 
					lcd_scroll_left();
					break;
				case 0xFB:
					lcd_scroll_right();
					break;
				case 0xFD:
					interpreter_program_counter -= 2;
					break;
				default:
					// TODO: invalid instruction.
					lcd_scroll_down(current_instruction & 0xF);
					break;
			}
			break;
		case 0x1000:
			interpreter_program_counter = current_instruction & 0x0FFF;
			break;
		case 0x2000:
			// TODO: stack overflow check.
			interpreter_stack[interpreter_stack_pointer++] = interpreter_program_counter;
			interpreter_program_counter = current_instruction & 0x0FFF;
			break;
		case 0x3000:
			if (INTERPRETER_REG_X == INTERPRETER_IMMEDIATE) interpreter_program_counter += 2;
			break;
		case 0x4000:
			if (INTERPRETER_REG_X != INTERPRETER_IMMEDIATE) interpreter_program_counter += 2;
			break;
		case 0x5000:
			if (INTERPRETER_REG_X == INTERPRETER_REG_Y) interpreter_program_counter += 2;
			break;
		case 0x6000:
			INTERPRETER_REG_X = INTERPRETER_IMMEDIATE;
			break;
		case 0x7000:
			INTERPRETER_REG_X += INTERPRETER_IMMEDIATE;
			break;
		case 0x8000:
			switch (current_instruction & 0x000F) {
				case 0x0:
					INTERPRETER_REG_X = INTERPRETER_REG_Y;
					break;
				case 0x1:
					INTERPRETER_REG_X |= INTERPRETER_REG_Y;
					break;
				case 0x2:
					INTERPRETER_REG_X &= INTERPRETER_REG_Y;
					break;
				case 0x3:
					INTERPRETER_REG_X ^= INTERPRETER_REG_Y;
					break;
				case 0x4:
					INTERPRETER_REG_F = (INTERPRETER_REG_X + INTERPRETER_REG_Y) > 0xFF;
					INTERPRETER_REG_X += INTERPRETER_REG_Y;
					break;
				case 0x5:
					INTERPRETER_REG_F = INTERPRETER_REG_X >= INTERPRETER_REG_Y;
					INTERPRETER_REG_X -= INTERPRETER_REG_Y;
					break;
				case 0x6:
					INTERPRETER_REG_F = INTERPRETER_REG_X & 0x01;
					INTERPRETER_REG_X >>= 1;
					break;
				case 0x7:
					INTERPRETER_REG_F = INTERPRETER_REG_Y >= INTERPRETER_REG_X;
					INTERPRETER_REG_X = INTERPRETER_REG_Y - INTERPRETER_REG_X;
					break;
				case 0xE:
					INTERPRETER_REG_F = (INTERPRETER_REG_X & 0x80) >> 7;
					INTERPRETER_REG_X <<= 1;
					break;
				default:
					// TODO: invalid instruction
					break;
			}
			break;
		case 0x9000:
			if (INTERPRETER_REG_X != INTERPRETER_REG_Y) interpreter_program_counter += 2;
			break;
		case 0xA000:
			interpreter_address_register = current_instruction & 0x0FFF;
			break;
		case 0xB000:
			interpreter_program_counter = (current_instruction & 0x0FFF) + interpreter_registers[0];
			break;
		case 0xC000:
			// TODO: random %
			INTERPRETER_REG_X = rand() & INTERPRETER_IMMEDIATE;
			break;
		case 0xD000:
			INTERPRETER_REG_F = 0;
			{
				uint8_t x, y, w, h, f;
				x = INTERPRETER_REG_X;
				y = INTERPRETER_REG_Y;
				w = 8;
				h = current_instruction & 0xF;
				f = 0;
				if (!(interpreter_flags & INTERPRETER_FLAGS_SCHIP)) {
					f |= LCD_SPRITE_FLAGS_ZOOMED;
					x *= 2;
					y *= 2;
				}
				
				if (h == 0) {
					if (interpreter_flags & INTERPRETER_FLAGS_SCHIP) {
						h = w = 16;
					} else {
						h = 16;
					}
				}
				
				if (interpreter_flags & INTERPRETER_FLAGS_SPRITES_WRAPPED_X) f |= LCD_SPRITE_FLAGS_WRAPPED_X;
				if (interpreter_flags & INTERPRETER_FLAGS_SPRITES_WRAPPED_Y) f |= LCD_SPRITE_FLAGS_WRAPPED_Y;
				
				INTERPRETER_REG_F = lcd_eor_sprite(
					interpreter_memory + interpreter_address_register - INTERPRETER_MEMORY_MIN,
					x, y, w, h, f
				) != 0;
			}
			break;
		case 0xE000:
			switch (current_instruction & 0xFF) {
				case 0x9E:
					if (keypad_get_key_state((INTERPRETER_REG_X) & 0xF, interpreter_flags)) interpreter_program_counter += 2;
					break;
				case 0xA1:
					if (!keypad_get_key_state((INTERPRETER_REG_X) & 0xF, interpreter_flags)) interpreter_program_counter += 2;
					break;
				default:
					// TODO: invalid instruction
					break;
			}
			break;
		case 0xF000:
			switch (current_instruction & 0xFF) {
				case 0x07:
					INTERPRETER_REG_X = interpreter_delay_timer;
					break;
				case 0x0A:
					#if KEYPAD_READ_KEY_IMPLEMENTED
						{
							uint8_t key = keypad_read_key();
							if (key && (interpreter_flags & INTERPRETER_FLAGS_KEYPAD_BLOCKING)) {
								while (keypad_get_key_state(key, interpreter_flags));
							}
							INTERPRETER_REG_X = key;
						}
					#else
						{
							uint8_t key_found = 0;
							for (uint8_t key = 0; key < 16; ++key) {
								if (keypad_get_key_state(key, interpreter_flags)) {
									INTERPRETER_REG_X = key;
									key_found = 1;
									break;
								}
							}
							if (!key_found) {
								interpreter_program_counter -= 2;
							} else if (interpreter_flags & INTERPRETER_FLAGS_KEYPAD_BLOCKING) {
								while (keypad_get_key_state(INTERPRETER_REG_X, interpreter_flags));
							}
						}
					#endif
					break;
				case 0x15:
					interpreter_delay_timer = INTERPRETER_REG_X;
					break;
				case 0x18:
					interpreter_sound_timer = INTERPRETER_REG_X;
					break;
				case 0x1E:
					interpreter_address_register += INTERPRETER_REG_X;
					break;
				case 0x29:
					interpreter_address_register = INTERPRETER_MEMORY_MIN + ((INTERPRETER_REG_X & 0xF) * 5);
					break;
				case 0x30:
					interpreter_address_register = INTERPRETER_MEMORY_MIN + ((INTERPRETER_REG_X & 0xF) * 10 + 5 * 16);
					break;
				case 0x33: {
					uint8_t to_bcd = INTERPRETER_REG_X;
					interpreter_memory[interpreter_address_register + 0 - INTERPRETER_MEMORY_MIN] = (to_bcd / 100);
					to_bcd -= interpreter_memory[interpreter_address_register + 0 - INTERPRETER_MEMORY_MIN] * 100;
					interpreter_memory[interpreter_address_register + 1 - INTERPRETER_MEMORY_MIN] = (to_bcd / 10);
					to_bcd -= interpreter_memory[interpreter_address_register + 1 - INTERPRETER_MEMORY_MIN] * 10;
					interpreter_memory[interpreter_address_register + 2 - INTERPRETER_MEMORY_MIN] = (to_bcd);
					} break;
				case 0x55:
					for (uint8_t i = 0; i <= (current_instruction & 0x0F00) >> 8; ++i) interpreter_memory[interpreter_address_register + i - INTERPRETER_MEMORY_MIN] = interpreter_registers[i];                                
					break;
				case 0x65:
					for (uint8_t i = 0; i <= (current_instruction & 0x0F00) >> 8; ++i) interpreter_registers[i] = interpreter_memory[interpreter_address_register + i - INTERPRETER_MEMORY_MIN];
					break;
				case 0x75:
					for (uint8_t i = 0; i <= (current_instruction & 0x0F00) >> 8; ++i) interpreter_hp48_flags[i] = interpreter_registers[i];
					break;
				case 0x85:
					for (uint8_t i = 0; i <= (current_instruction & 0x0F00) >> 8; ++i) interpreter_registers[i] = interpreter_hp48_flags[interpreter_address_register];
					break;
				default:
					// TODO: invalid instruction
					break;
			}
			break;
		default:
			// TODO: invalid instruction
			break;
	}
}

// Run the CHIP-8 timers for one "tick". Should be called 60 times a second.
void interpreter_tick(void) {
	if (interpreter_delay_timer) --interpreter_delay_timer;
	if (interpreter_sound_timer) --interpreter_sound_timer;
}