#include <stdint.h>

#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "interpreter.h"
#include "lcd.h"
#include "keypad.h"
#include "ui.h"

#include "games.h"

// Minimum game speed is defined speed/MIN_SPEED_DIVISOR
#define MIN_SPEED_DIVISOR    (8)
// maximum game speed is defined speed*MAX_SPEED_MULTIPLIER
#define MAX_SPEED_MULTIPLIER (8)

// Number of pixels to scroll the game list at a time. Must be a factor of 8.
#define GAME_LIST_SCROLL_SPEED (4)

volatile int32_t instructions_to_execute = 0;
uint32_t interpreter_speed = 0;

ISR(TIMER2_COMPA_vect) {
	// Run the CHIP-8 timers for one tick.
	interpreter_tick();
	// If the sound timer is non-zero, make a noise.
	if (interpreter_sound_timer) {
		DDRB |= _BV(3);
	} else {
		DDRB &= (uint8_t)~_BV(3);
	}
	
	// Request a number of instructions to execute.
	
	// Calculate the speed of the interpreter.
	uint8_t speed_dial = ADCH;
	if (speed_dial < 128) {
		instructions_to_execute = (interpreter_speed * 127) / (127 + ((MIN_SPEED_DIVISOR - 1) * (127 - speed_dial)));
	} else {
		instructions_to_execute = interpreter_speed + (interpreter_speed * (speed_dial - 128) * (MAX_SPEED_MULTIPLIER - 1)) / 127;
	}	
}

const uint8_t* get_game(uint8_t game_index) {
	const uint8_t* p;
	memcpy_P(&p, &games[game_index], sizeof(const uint8_t*));
	return p;
}

// Retrieves the offset to a game's name.
const char* get_game_name(uint8_t game_index) {
	const uint8_t* p = (get_game(game_index) + 2 + 1 + 8);
	p += (pgm_read_byte(p + 0) + (pgm_read_byte(p + 1) << 8) + 2);
	return (const char*)p;
}

// Retrieves the offset to a game's description string from its index.
const char* get_game_description(uint8_t game_index) {
	const char* p = get_game_name(game_index);
	while (pgm_read_byte(p++));
	return p;
}

// Loads a game from its index.
void load_game(uint8_t game_index) {

	
	// Start by resetting the interpreter.
	interpreter_reset();
	
	// Get the offset to the game data.
	const uint8_t* game = get_game(game_index);
	
	// Set the speed.
	interpreter_speed = pgm_read_byte(game + 0) + (pgm_read_byte(game + 1) << 8);
	if (interpreter_speed == 0) interpreter_speed = 8000;
	game += 2;
	
	// Set the flags.
	interpreter_flags = pgm_read_byte(game);
	game += 1;
	
	// Skip the keys.
	game += 8;
	
	// Get the size of the game.
	uint16_t game_size = pgm_read_byte(game + 0) + (pgm_read_byte(game + 1) << 8);
	game += 2;
	
	// Load the game.
	interpreter_load_P(game, game_size);
	
	instructions_to_execute = 0;
}

// Waits for a confirmation/fire key to be pressed ([5] or [F]).
void wait_confirm(void) {
	// Wait for the fire keys to be pressed.
	while (!(keypad_get_key_state(0x5, 0) || keypad_get_key_state(0xF, 0)));
	// Wait for the fire keys to be released.
	while (keypad_get_key_state(0x5, 0) || keypad_get_key_state(0xF, 0));
}

// Draws a game selection menu and returns the index of the selected game.
uint8_t game_selection_menu(void) {
	
	// Store the current and previous game indices.
	uint8_t game_index = 0, previous_game_index = 0xFF;
	// Store the current and previous game list scroll position.
	uint8_t game_list_scroll = 0, previous_game_list_scroll = 0xFF;

	for(;;) {
	
		// Do we need to repaint the entire game selection menu?
		if (game_list_scroll != previous_game_list_scroll) {
			
			uint8_t x = 0, y = 0;
			const char* name;
			
			if (previous_game_list_scroll == 0xFF) {
			
				// Full redraw required.

				lcd_clear();
				
				for (uint8_t menu_game_index = 0; menu_game_index < 8; ++menu_game_index) {			
					name = get_game_name(game_list_scroll + menu_game_index);
					x = (128 - ui_measure_string_width_P(name)) / 2;
					ui_put_string_P(&x, &y, name);
					y += 8;
				}
				
				// We don't have any arrows to erase.
				previous_game_index = 0xFF;
			
			} else if (game_list_scroll > previous_game_list_scroll) {
			
				// Scrolling down.
				for (uint8_t r = 0; r < 8; r += GAME_LIST_SCROLL_SPEED) lcd_scroll_up(GAME_LIST_SCROLL_SPEED);
				
				// Draw bottom item.
				name = get_game_name(game_list_scroll + 7);
				x = (128 - ui_measure_string_width_P(name)) / 2;
				y = 64 - 8;
				ui_put_string_P(&x, &y, name);
			
			} else if (game_list_scroll < previous_game_list_scroll) {
				
				// Scrolling up.
				for (uint8_t r = 0; r < 8; r += GAME_LIST_SCROLL_SPEED) lcd_scroll_down(GAME_LIST_SCROLL_SPEED);
				
				// Draw top item.
				name = get_game_name(game_list_scroll);
				x = (128 - ui_measure_string_width_P(name)) / 2;
				y = 0;
				ui_put_string_P(&x, &y, name);
			}
			
			previous_game_list_scroll = game_list_scroll;
		}
		
		// Do we need to redraw the game selection marker?
		if (game_index != previous_game_index) {
			uint8_t x = 0, y = 0;
			// Erase the old marker if required.
			if (previous_game_index != 0xFF) {
				y = (previous_game_index - game_list_scroll) * 8;
				x = -1;
				ui_put_char(&x, &y, UI_CHAR_ARROW_RIGHT);
				x = 128 - 7;
				ui_put_char(&x, &y, UI_CHAR_ARROW_LEFT);
			}
			// Draw the new marker.
			y = (game_index - game_list_scroll) * 8;
			x = -1;
			ui_put_char(&x, &y, UI_CHAR_ARROW_RIGHT);
			x = 128 - 7;
			ui_put_char(&x, &y, UI_CHAR_ARROW_LEFT);
			previous_game_index = game_index;
		}
		
		// Check for fire key (5/F).
		if (keypad_get_key_state(0x5, 0) || keypad_get_key_state(0xF, 0)) break;
		// Check for up key.
		if (keypad_get_key_state(0x2, 0)) {
			game_index = (game_index + GAME_COUNT-1) % GAME_COUNT;
			while (keypad_get_key_state(0x2, 0));
			if (game_index == GAME_COUNT - 1) {
				game_list_scroll = GAME_COUNT - 8;
				previous_game_list_scroll = 0xFF;
			} else if (game_index < game_list_scroll) {
				--game_list_scroll;
			}
		}
		// Check for down key.
		if (keypad_get_key_state(0x8, 0)) {
			game_index = (game_index + 1) % GAME_COUNT;
			if (game_index == 0) {
				game_list_scroll = 0;
				previous_game_list_scroll = 0xFF;
			} else if (game_index >= game_list_scroll + 8) {
				++game_list_scroll;
			}
			while (keypad_get_key_state(0x8, 0));
		}
	}
	// Wait for the fire keys to be released.
	while (keypad_get_key_state(0x5, 0) || keypad_get_key_state(0xF, 0));	
	
	return game_index;
}


int main(void) {

	// Initialise the LCD.
	lcd_init();
	lcd_clear();
	lcd_display_on();
	
	// Initialise the keypad.
	keypad_init();
	
	// Initialise the interpreter.
	interpreter_reset();
	
	// Use timer 0 to generate sound output.
	TCCR0A |= _BV(COM0A0) | _BV(WGM01);
	TCCR0B |= _BV(WGM02) | (0b101 << CS00); // clkIO/1024
	OCR0A = (F_CPU / 1024 / (440 * 2)) - 1; // A440

	// Use timer 2 for CHIP-8 delay timers.
	TCCR2B |= (0b111 << CS20);       // clkIO/1024
	OCR2A = (F_CPU / 1024 / 60) - 1; // Timers are 60Hz
	TIMSK2 |= _BV(OCIE2A);	         // Generate an interrupt
	
	// Enable interrupts to update the timers.
	sei();
	
	// Configure the ADC.
	ADMUX |= _BV(REFS0) | _BV(ADLAR); // AVCC reference, left-aligned results, ADC0 single-ended input.
	ADCSRA |=  _BV(ADEN) | _BV(ADSC) | _BV(ADATE) | _BV(ADPS2) | _BV(ADPS1); // Enable ADC, start conversion, auto-trigger (free running), prescaler = 128 (8MHz/128=125kHz).
	DIDR0 |= _BV(0); // Disable digital input 0.

	// Show the game selection menu.
	uint8_t game_index = game_selection_menu();
	
	// Show the game description.
	lcd_clear();
	uint8_t description_x = 0;
	uint8_t description_y = 0;
	ui_put_wrapped_string_P(&description_x, &description_y, get_game_description(game_index), 96, 8);
	
	// Show the game keys.
	const uint8_t* keys = get_game(game_index) + 2 /* <- speed */;
	uint8_t flags = pgm_read_byte(keys++);
	
	for (uint8_t y = 0, k = 0; y < 32; y += 8) {
		for (uint8_t x = 0; x < 32; x += 8, ++k) {
			uint8_t cx = x + 97;
			uint8_t cy = 33 + ((flags & INTERPRETER_FLAGS_KEYPAD_FLIPPED_Y) ? (24 - y) : y);
			ui_put_char(&cx, &cy, UI_CHAR_BLOCK);
			uint8_t c = pgm_read_byte(keys + (k / 2));
			if (k & 1) c >>= 4;
			c &= 0xF;
			if (c != 0) {
				cx -= 8;
				ui_put_char(&cx, &cy, UI_CHAR_BLOCK + c);
			}
		}
	}
	wait_confirm();

	// Load the game into the interpreter.
	load_game(game_index);	

	// Run the interpreter.
	for(;;) {
		while (instructions_to_execute > 0) {
			interpreter_execute_instruction();
			cli();
			instructions_to_execute -= 60;
			sei();
		}
	}
}