#include <stdint.h>
#include <avr/io.h>
#include <util/delay.h>

#include "lcd.h"
#include "pg12864lrs-ann-h.h"
#include "string.h"

// Initialise the LCD.
void lcd_init(void) {
	
	// Leave data port floating.
	LCD_DATA_DDR = 0x00;
	LCD_DATA_PORT = 0x00;
	
	// Make the control pins outputs.
	LCD_CONTROL_DDR |= _BV(LCD_REGISTER_SELECT) | _BV(LCD_READ_WRITE) | _BV(LCD_ENABLE) | _BV(LCD_CS1) | _BV(LCD_CS2) | _BV(LCD_RESET);
	// Make LCD_REGISTER_SELECT, LCD_ENABLE and LCD_RESET high.
	LCD_CONTROL_PORT |= _BV(LCD_REGISTER_SELECT) | _BV(LCD_ENABLE) | _BV(LCD_RESET);
	// Make LCD_READ_WRITE, LCD_CS1 and LCD_CS2 low.
	LCD_CONTROL_PORT &= (uint8_t)~(_BV(LCD_READ_WRITE) | _BV(LCD_CS1) | _BV(LCD_CS2));

	// Pulse the reset pin.
	_delay_ms(20);
	LCD_CONTROL_PIN |= _BV(LCD_RESET);
	_delay_ms(20);
	LCD_CONTROL_PIN |= _BV(LCD_RESET);
}

// Select parts of the LCD. You should pass a bit mask of _BV(LCD_CS1) and _BV(LCD_CS2).
void lcd_select(uint8_t parts) {
	parts &= (uint8_t)(_BV(LCD_CS1) | _BV(LCD_CS2));
	switch (parts) {
		case 0:
			LCD_CONTROL_PORT &= (uint8_t)~(_BV(LCD_CS1) | _BV(LCD_CS2));
			break;
		case _BV(LCD_CS1):
		case _BV(LCD_CS2):
			LCD_CONTROL_PORT = (LCD_CONTROL_PORT & (uint8_t)~(_BV(LCD_CS1) | _BV(LCD_CS2))) | parts;
			break;
		case _BV(LCD_CS1) | _BV(LCD_CS2):
			LCD_CONTROL_PORT |= _BV(LCD_CS1) | _BV(LCD_CS2);
			break;
	}
}

// Write a byte to the LCD. CSn and REGISTER_SELECT pins should have been configured before calling this function.
void lcd_write(uint8_t value) {
	// Put LCD_ENABLE and LCD_READ_WRITE low.
	LCD_CONTROL_PORT &= (uint8_t)~(_BV(LCD_ENABLE) | _BV(LCD_READ_WRITE));
	// Delay for at least 450ns.
	_delay_us(1);
	// Put LCD_ENABLE high.
	LCD_CONTROL_PORT |= _BV(LCD_ENABLE);
	// Put the data on the data bus.
	LCD_DATA_DDR = 0xFF;
	LCD_DATA_PORT = value;
	// Delay for at least 450ns.
	_delay_us(1);
	// Put LCD_ENABLE low.
	LCD_CONTROL_PORT &= (uint8_t)~_BV(LCD_ENABLE);
	// Release the data bus.
	LCD_DATA_DDR = 0x00;
	LCD_DATA_PORT = 0x00;
	// Delay for at least 10ns.
	_delay_us(1);
}

// Write a data byte to the LCD. CSn pins should have been configured before calling this function.
void lcd_write_data(uint8_t value) {
	LCD_CONTROL_PORT |= _BV(LCD_REGISTER_SELECT);
	lcd_write(value);
}

// Write an instruction byte to the LCD. CSn pins should have been configured before calling this function.
void lcd_write_instruction(uint8_t value) {
	LCD_CONTROL_PORT &= (uint8_t)~_BV(LCD_REGISTER_SELECT);
	lcd_write(value);
}

// Read a byte from the LCD. CSn and REGISTER_SELECT pins should have been configured before calling this function.
uint8_t lcd_read(void) {
	uint8_t result;
	// Release the data bus.
	LCD_DATA_DDR = 0x00;
	LCD_DATA_PORT = 0x00;
	// Put LCD_ENABLE low.
	LCD_CONTROL_PORT &= (uint8_t)~_BV(LCD_ENABLE);
	// Put LCD_READ_WRITE high.
	LCD_CONTROL_PORT |= _BV(LCD_READ_WRITE);
	// Delay for at least 450ns.
	_delay_us(1);
	// Put LCD_ENABLE high.
	LCD_CONTROL_PORT |= _BV(LCD_ENABLE);
	// Wait at least 320nS.
	_delay_us(1);
	// Read the data.
	result = LCD_DATA_PIN;
	// Put LCD_ENABLE low.
	LCD_CONTROL_PORT &= (uint8_t)~_BV(LCD_ENABLE);
	// Delay for at least 20ns.
	_delay_us(1);
	return result;
}

// Read a data byte from the LCD. CSn pins should have been configured before calling this function.
uint8_t lcd_read_data(void) {
	LCD_CONTROL_PORT |= _BV(LCD_REGISTER_SELECT);
	return lcd_read();
}

// Read an instruction byte from the LCD. CSn pins should have been configured before calling this function.
uint8_t lcd_read_instruction(void) {
	LCD_CONTROL_PORT &= (uint8_t)~_BV(LCD_REGISTER_SELECT);
	return lcd_read();
}

// Reads the LCD and returns non-zero if the LCD is busy, zero otherwise.
uint8_t lcd_busy(void) {
	return lcd_read_instruction() & 0x80;
}

// Switch the display on.
void lcd_display_on(void) {
	lcd_select(_BV(LCD_CS1));
	while (lcd_busy());
	lcd_write_instruction(LCD_INSTRUCTION_DISPLAY_ON);
	lcd_select(_BV(LCD_CS2));
	while (lcd_busy());
	lcd_write_instruction(LCD_INSTRUCTION_DISPLAY_ON);
}

// Switch the display off.
void lcd_display_off(void) {
	lcd_select(_BV(LCD_CS1));
	while (lcd_busy());
	lcd_write_instruction(LCD_INSTRUCTION_DISPLAY_OFF);
	lcd_select(_BV(LCD_CS2));
	while (lcd_busy());
	lcd_write_instruction(LCD_INSTRUCTION_DISPLAY_OFF);
}

// Set the address (horizontal coordinate). LCD_CS1 and LCD_CS2 should be configured beforehand.
void lcd_set_address(uint8_t address) {
	address &= 0b111111;
	while (lcd_busy());
	lcd_write_instruction(LCD_INSTRUCTION_SET_ADDRESS | address);
}

// Set the page (vertical coordinate). LCD_CS1 and LCD_CS2 should be configured beforehand.
void lcd_set_page(uint8_t page) {
	while (lcd_busy());
	lcd_write_instruction(LCD_INSTRUCTION_SET_PAGE | (page & 0b111));
}

// Set the start line. LCD_CS1 and LCD_CS2 should be configured beforehand.
void lcd_set_start_line(uint8_t start_line) {
	while (lcd_busy());
	lcd_write_instruction(LCD_INSTRUCTION_DISPLAY_START_LINE | (start_line & 0b111111));
}

// Clear the LCD.
void lcd_clear(void) {
	uint8_t cs, address, page;
	for (cs = 0; cs < 2; ++cs) {
		if (cs == 0) {
			lcd_select(_BV(LCD_CS1));
		} else {
			lcd_select(_BV(LCD_CS2));
		}
		for (page = 0; page < 8; ++page) {
			// Set the address.
			while (lcd_busy());
			lcd_set_address(0);
			// Set the start line (if on the first page).
			if (page == 0) {
				while (lcd_busy());
				lcd_set_start_line(0);
			}
			// Set the page.
			while (lcd_busy());
			lcd_set_page(page);
			// Write 64 bytes to clear the current page.
			for (address = 0; address < 64; ++address) {
				while (lcd_busy());
				lcd_write_data(0x00);
			}
		}
	}
}

// Invert a pixel on the LCD.
uint8_t lcd_invert_pixel(uint8_t x, uint8_t y) {

	uint8_t old_value, pixel_mask;

	// Wrap the pixel coordinates.
	x &= 127;
	y &= 63;

	pixel_mask = _BV(y & 7);

	if (x < 64) {
		lcd_select(_BV(LCD_CS1));
	} else {
		lcd_select(_BV(LCD_CS2));
	}

	// Set the address.
	while (lcd_busy());
	lcd_set_address(x);

	// Set the page.
	while (lcd_busy());
	lcd_set_page(y / 8);

	// Dummy read.
	while (lcd_busy());
	lcd_read_data();

	// Actual read.
	while (lcd_busy());
	old_value = lcd_read_data();

	// Re-set the address.
	while (lcd_busy());
	lcd_set_address(x);

	// Write the new pixel.
	while (lcd_busy());
	lcd_write_data(old_value ^ pixel_mask);

	// Return the collision value.
	return pixel_mask & old_value;
}
// Draws a sprite on the LCD.
uint8_t lcd_eor_sprite(uint8_t* data, int8_t x, int8_t y, uint8_t width, uint8_t height, uint8_t flags) {
	
	uint8_t collision;
	uint8_t sub_x, sub_y;
	uint8_t byte_width, width_in_bytes, byte_x, byte_data;
	uint8_t total_height;
	uint8_t dest_x, dest_y;

	// Stores the collision mask.
	collision = 0;

	// Width of the sprite in whole bytes.
	width_in_bytes = (width + 7) / 8;

	// How wide is a byte of the source going to appear on the screen?
	if (flags & LCD_SPRITE_FLAGS_ZOOMED) {
		byte_width = 16;
		total_height = height * 2;
	} else {
		byte_width = 8;
		total_height = height;
	}

	for (sub_y = 0, dest_y = y; sub_y < total_height; ++sub_y, ++dest_y) {
		for (byte_x = 0, dest_x = x; byte_x < width_in_bytes; ++byte_x) {
			byte_data = *(data++);
			for (sub_x = 0; sub_x < byte_width; ++sub_x, ++dest_x) {
				if ((byte_data & 0x80) && ((flags & LCD_SPRITE_FLAGS_WRAPPED_X) || (dest_x >= 0 && dest_x < 128)) && ((flags & LCD_SPRITE_FLAGS_WRAPPED_Y) || (dest_y >= 0 && dest_y < 64))) {
					collision |= lcd_invert_pixel(dest_x, dest_y);
				}
				if (!(flags & LCD_SPRITE_FLAGS_ZOOMED) || (sub_x & 1)) {
					byte_data <<= 1;
				}
			}
		}
		if ((flags & LCD_SPRITE_FLAGS_ZOOMED) && !(sub_y & 1)) data -= width_in_bytes;
	}

	return collision;

}

uint8_t scroll_buffer[128] __attribute__ ((section (".noinit")));

// Scroll the LCD down 0..7 lines.
void lcd_scroll_down_partial(uint8_t lines) {
	uint8_t page, address, half;
	uint8_t *previous_row, *current_row;
	
	for (half = 0; half < 128; half += 64) {
		
		previous_row = scroll_buffer + 0;
		current_row = scroll_buffer + 64;
		
		memset(previous_row, 0x00, 64);
		
		if (half == 0) {
			lcd_select(_BV(LCD_CS1));
		} else {
			lcd_select(_BV(LCD_CS2));
		}		
		
		for (page = 0; page < 8; ++page) {		
			
			// Set the address.
			while (lcd_busy());
			lcd_set_address(0);
			
			// Set the page.
			while (lcd_busy());
			lcd_set_page(page);
			
			// Dummy read.
			while (lcd_busy());
			lcd_read_data();
			
			// Read half of the row.
			for (address = 0; address < 64; ++address) {
				uint8_t current;
				while (lcd_busy());
				current = lcd_read_data();
				current_row[address] = (current << lines) | (previous_row[address] >> (8 - lines));
				previous_row[address] = current;
			}
			
			// Write back the manipulated row.
			while (lcd_busy());
			lcd_set_address(0);
			for (address = 0; address < 64; ++address) {
				while (lcd_busy());
				lcd_write_data(current_row[address]);
			}			
		}
	}
}

// Scroll the LCD down 0..15 lines.
void lcd_scroll_down(uint8_t lines) {
	while (lines > 7) {
		lcd_scroll_down_partial(8);
		lines -= 8;
	}
	lcd_scroll_down_partial(lines);
}


// Scroll the LCD up 0..7 lines.
void lcd_scroll_up_partial(uint8_t lines) {
	int8_t page;
	uint8_t address, half;
	uint8_t *previous_row, *current_row;
	
	for (half = 0; half < 128; half += 64) {
		
		previous_row = scroll_buffer + 0;
		current_row = scroll_buffer + 64;
		
		memset(previous_row, 0x00, 64);
		
		if (half == 0) {
			lcd_select(_BV(LCD_CS1));
		} else {
			lcd_select(_BV(LCD_CS2));
		}		
		
		for (page = 7; page >= 0; --page) {		
			
			// Set the address.
			while (lcd_busy());
			lcd_set_address(0);
			
			// Set the page.
			while (lcd_busy());
			lcd_set_page(page);
			
			// Dummy read.
			while (lcd_busy());
			lcd_read_data();
			
			// Read half of the row.
			for (address = 0; address < 64; ++address) {
				uint8_t current;
				while (lcd_busy());
				current = lcd_read_data();
				current_row[address] = (current >> lines) | (previous_row[address] << (8 - lines));
				previous_row[address] = current;
			}
			
			// Write back the manipulated row.
			while (lcd_busy());
			lcd_set_address(0);
			for (address = 0; address < 64; ++address) {
				while (lcd_busy());
				lcd_write_data(current_row[address]);
			}			
		}
	}
}

// Scroll the LCD up 0..15 lines.
void lcd_scroll_up(uint8_t lines) {
	while (lines > 7) {
		lcd_scroll_up_partial(8);
		lines -= 8;
	}
	lcd_scroll_up_partial(lines);
}

// Scroll the LCD left four columns.
void lcd_scroll_left(void) {
	uint8_t page, address;
	int8_t half;
	
	for (page = 0; page < 8; ++page) {	
		for (half = 64; half >= 0; half -= 64) {
			
			if (half == 0) {
				lcd_select(_BV(LCD_CS1));
			} else {
				lcd_select(_BV(LCD_CS2));
			}
			
			// Read half of the row.
			if (half == 0) {
				address = 4;
			} else {
				address = 0;
			}
			
			// Set the address.
			while (lcd_busy());
			lcd_set_address(address);
			
			// Set the page.
			while (lcd_busy());
			lcd_set_page(page);
			
			// Dummy read.
			while (lcd_busy());
			lcd_read_data();

			for (; address < 64; ++address) {
				while (lcd_busy());
				scroll_buffer[address] = lcd_read_data();
			}
			
			// Write back the row.
			while (lcd_busy());
			lcd_set_address(0);
			for (address = 0; address < 60; ++address) {
				while (lcd_busy());
				lcd_write_data(scroll_buffer[address + 4]);
			}
			
			if (half == 0) { 
				// Finish off with four from right hand side of the display.
				for (; address < 64; ++address) {
					while (lcd_busy()); 
					lcd_write_data(scroll_buffer[address - 60]);
				}
			} else {
				// Finish off with four blanks.
				for (; address < 64; ++address) {
					while (lcd_busy()); 
					lcd_write_data(0x00);
				}
			}
		}
	}
}

// Scroll the LCD right four columns.
void lcd_scroll_right(void) {
	uint8_t page, address, max_address, half;
	
	for (page = 0; page < 8; ++page) {	
		for (half = 0; half < 128; half += 64) {
			
			if (half == 0) {
				lcd_select(_BV(LCD_CS1));
				max_address = 64;
			} else {
				lcd_select(_BV(LCD_CS2));
				max_address = 60;
			}
			
			// Set the address.
			while (lcd_busy());
			lcd_set_address(0);
			
			// Set the page.
			while (lcd_busy());
			lcd_set_page(page);
					
			// Dummy read.
			while (lcd_busy());
			lcd_read_data();

			// Read the current data.
			for (address = 0; address < max_address; ++address) {
				while (lcd_busy());
				scroll_buffer[address] = lcd_read_data();
			}
			
			// Set the address.
			while (lcd_busy());
			lcd_set_address(0);
			
			// Write the first four columns.
			if (half == 0) {
				for (address = 0; address < 4; ++address) {
					while (lcd_busy());
					lcd_write_data(0x00);
				}
			} else {
				for (address = 0; address < 4; ++address) {
					while (lcd_busy());
					lcd_write_data(scroll_buffer[address + 60]);
				}
			}
			
			// Write back the rest of the row.
			for (; address < 64; ++address) {
				while (lcd_busy());
				lcd_write_data(scroll_buffer[address - 4]);
			}
		}
	}
}