﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Text;

namespace VinegarTools.GameEncoder {
	class Program {

		class Game {
			public string Name { get; set; }
			public string Description { get; set; }
			public ushort Speed { get; set; }
			public string FlagsOn { get; set; }
			public string FlagsOff { get; set; }
			public byte[] Data { get; set; }
			public string Keys { get; set; }

			public string GetFlagsString() {
				var result = "(INTERPRETER_FLAGS_DEFAULT";
				if (!string.IsNullOrEmpty(this.FlagsOn)) {
					result += " | " + string.Join(" | ", Array.ConvertAll(this.FlagsOn.Split(','), s => "INTERPRETER_FLAGS_" + s));
				}
				result += ")";
				if (!string.IsNullOrEmpty(this.FlagsOff)) {
					result += " & (uint8_t)~(" + string.Join(" | ", Array.ConvertAll(this.FlagsOff.Split(','), s => "INTERPRETER_FLAGS_" + s)) + ")";
				}
				return result;
			}
		}

		static int MapKeyCode(int key) {
			switch (key & 0x0F) {
				case 0x0: return 0xD;
				case 0x1: return 0x0;
				case 0x2: return 0x1;
				case 0x3: return 0x2;
				case 0x4: return 0x4;
				case 0x5: return 0x5;
				case 0x6: return 0x6;
				case 0x7: return 0x8;
				case 0x8: return 0x9;
				case 0x9: return 0xA;
				case 0xA: return 0xC;
				case 0xB: return 0xE;
				case 0xC: return 0x3;
				case 0xD: return 0x7;
				case 0xE: return 0xB;
				case 0xF: return 0xF;
			}
			return 0;
		}

		static void Main(string[] args) {

			// Cap the number of games.
			int gameLimit = 0;

			// Enable to ignore the game data, which produces a far smaller binary (useful to test menus etc).
			bool datalessGames = false;

			// Build a list of games.
			var games = new List<Game>();

			// Parse the INI file.
			Game currentGame = null;
			foreach (var sourceIniLine in File.ReadAllLines("../../../../games/games.ini")) {
				var iniLine = sourceIniLine.Split('#')[0].Trim();
				if (iniLine == "") continue;
				if (iniLine.StartsWith("[") && iniLine.EndsWith("]")) {
					var gameName = iniLine.Substring(1, iniLine.Length - 2).Trim();
					var gameSourceFilePath = Path.Combine("../../../../games", gameName);
					currentGame = new Game() {
						Name = Path.GetFileNameWithoutExtension(gameSourceFilePath),
						Data = datalessGames ? new byte[0] : File.ReadAllBytes(gameSourceFilePath),
					};
					games.Add(currentGame);
				} else {
					var equalsPos = iniLine.IndexOf('=');
					if (equalsPos < 0) throw new InvalidDataException("Missing '='.");
					var iniKey = iniLine.Substring(0, equalsPos).TrimEnd();
					var iniValue = iniLine.Substring(equalsPos + 1).TrimStart();
					switch (iniKey.ToLowerInvariant()) {
						case "name":
							currentGame.Name = iniValue;
							break;
						case "description":
							currentGame.Description = iniValue;
							break;
						case "speed":
							currentGame.Speed = ushort.Parse(iniValue, CultureInfo.InvariantCulture);
							break;
						case "flagsoff":
							currentGame.FlagsOff = iniValue;
							break;
						case "flagson":
							currentGame.FlagsOn = iniValue;
							break;
						case "keys":
							currentGame.Keys = iniValue;
							break;
						default:
							throw new InvalidDataException(string.Format("Invalid key {0}.", iniKey));
					}
				}
			}

			// Sort the games in name order.
			games.Sort((a, b) => a.Name.CompareTo(b.Name));

			// If there's a limit, pick the smallest games.
			if (gameLimit > 0) {
				games.Sort((a, b) => a.Data.Length.CompareTo(b.Data.Length));
				while (games.Count > gameLimit) games.RemoveAt(gameLimit);
			}

			// Write the output file.
			using (var outputFile = new StreamWriter(File.Create(args[0]))) {

				// Start.
				outputFile.WriteLine("#ifndef _GAMES_H_");
				outputFile.WriteLine("#define _GAMES_H_");
				outputFile.WriteLine();
				outputFile.WriteLine("#include <stdint.h>");
				outputFile.WriteLine("#include <avr/pgmspace.h>");
				outputFile.WriteLine("#include \"ui.h\"");
				outputFile.WriteLine();
				outputFile.WriteLine("#define GAME_COUNT ({0})", games.Count);
				outputFile.WriteLine();

				// Output data.
				for (int gameIndex = 0; gameIndex < games.Count; ++gameIndex) {

					var game = games[gameIndex];

					outputFile.WriteLine("const uint8_t game_{0}[] PROGMEM = {{ /* {1} */", gameIndex, game.Name);

					// Speed
					outputFile.WriteLine("\t(uint8_t){0}, (uint8_t)({0} >> 8),", game.Speed);

					// Flags
					outputFile.WriteLine("\t{0},", game.GetFlagsString());

					// Keys
					var assignedKeys = new string[16];
					for (int i = 0; i < assignedKeys.Length; ++i) assignedKeys[i] = "0x00";
					if (!string.IsNullOrEmpty(game.Keys)) {
						foreach (var keyDefinition in game.Keys.Split(',')) {
							var keyDefintionSplit = keyDefinition.IndexOf(':');
							if (keyDefintionSplit < 0) throw new InvalidDataException("Missing ':'.");
							var keyId = int.Parse(keyDefinition.Substring(0, keyDefintionSplit).TrimEnd(), NumberStyles.HexNumber);

							var keyName = keyDefinition.Substring(keyDefintionSplit + 1).ToUpperInvariant();
							switch (keyName) {
								case "UP":
								case "DOWN":
								case "LEFT":
								case "RIGHT":
									keyName = "ARROW_" + keyName;
									break;
							}
							assignedKeys[MapKeyCode(keyId)] = "UI_CHAR_" + keyName + "-UI_CHAR_BLOCK";
						}
					}
					outputFile.Write("\t");
					for (int i = 0; i < assignedKeys.Length; i += 2) {
						outputFile.Write("(uint8_t)(({0})|(({1})<<4)), ", assignedKeys[i + 0], assignedKeys[i + 1]);
					}
					outputFile.WriteLine();

					// Size
					outputFile.WriteLine("\t(uint8_t){0},  (uint8_t)({0} >> 8),", game.Data.Length);

					// Actual data.
					for (int i = 0; i < game.Data.Length; ++i) {
						if ((i % 16) == 0) outputFile.Write("\t");
						outputFile.Write("0x{0:X2}, ", game.Data[i]);
						if ((i % 16) == 15) outputFile.WriteLine();
					}
					if ((game.Data.Length % 16) != 0) outputFile.WriteLine();

					// Name
					outputFile.Write("\t");
					foreach (var dataByte in Encoding.ASCII.GetBytes(game.Name ?? "")) outputFile.Write("0x{0:X2}, ", dataByte);
					outputFile.WriteLine("0x00,");

					// Description
					outputFile.Write("\t");
					foreach (var dataByte in Encoding.ASCII.GetBytes(game.Description ?? "")) outputFile.Write("0x{0:X2}, ", dataByte);
					outputFile.WriteLine("0x00,");

					outputFile.WriteLine("};");
					outputFile.WriteLine();
	
				}

				// Game list.
				outputFile.WriteLine("const uint8_t* games[{0}] PROGMEM = {{", games.Count);
				for (int i = 0; i < games.Count; ++i) {
					outputFile.WriteLine("\tgame_{0}, /* {1} */", i, games[i].Name);
				}
				outputFile.WriteLine("};");
				outputFile.WriteLine();

				// Finish.
				outputFile.WriteLine("#endif");
			}
		}
	}
}
