#include <string.h>

#include "lcd.h"
#include "ui_font.h"
#include "ui.h"


uint8_t expanded_character[8];

// Draw a character at a particular X/Y coordinate. X and Y are updated by the function.
void ui_put_char(uint8_t* x, uint8_t* y, char c) {
	
	uint16_t data_offset;
	uint8_t char_width, i, j, char_data;
	
	// Clear the expanded character buffer.
	memset(expanded_character, 0, 8);
	
	// Subtract 32 as our first character is number 32.
	c -= 32;
	
	// Calculate the offset to the character data in the font.
	data_offset = pgm_read_byte(ui_font + c * 2) + (pgm_read_byte(ui_font + c * 2 + 1) << 8);
	
	// Get the width of the character.
	char_width = pgm_read_byte(ui_font + data_offset);
	
	// Rotate the character, starting from the last column.
	data_offset += char_width;
	for (i = 0; i < char_width; ++i) {
		// Pull in the column.
		char_data = pgm_read_byte(ui_font + data_offset--);
		// Split the bits in the column into rows.
		for (j = 0; j < 8; ++j) {
			expanded_character[j] >>= 1;
			expanded_character[j] |= (char_data & 0x80);
			char_data <<= 1;
		}
	}
	
	// Draw the character as a prite.
	lcd_eor_sprite(expanded_character, *x, *y, char_width, 8, 0);
	*x += char_width;
}

// Draw a string on the screen at a particular X/Y coordinate. X and Y are updated by the function.
void ui_put_string(uint8_t* x, uint8_t* y, const char* s) {
	char c;
	while ((c = *(s++))) ui_put_char(x, y, c);
}

// Draw a string stored in ROM on the screen at a particular X/Y coordinate. X and Y are updated by the function.
void ui_put_string_P(uint8_t* x, uint8_t* y, const char* s) {
	char c;
	while ((c = pgm_read_byte(s++))) ui_put_char(x, y, c);
}

// Measures the width of a string stored in ROM.
uint8_t ui_measure_string_width_P(const char* s) {
	char c;
	uint8_t total_width;
	uint16_t data_offset;
	
	total_width = 0;
	while ((c = pgm_read_byte(s++))) {
		// Subtract 32 as our first character is number 32.
		c -= 32;
	
		// Calculate the offset to the character data in the font.
		data_offset = pgm_read_byte(ui_font + c * 2) + (pgm_read_byte(ui_font + c * 2 + 1) << 8);
	
		// Get the width of the character.
		total_width += pgm_read_byte(ui_font + data_offset);
	}
	return total_width;
}


// Draws a string wrapping it at particular points.
void ui_put_wrapped_string_P(uint8_t* x, uint8_t* y, const char* s, uint8_t width, uint8_t line_height) {
	uint8_t original_x = *x;
	uint8_t width_remaining = width;
	uint8_t word_width;
	const char* t;
	char c;
	uint16_t data_offset;

	do {
		
		// Preserve the start of the current word.
		t = s;
		// Measure the word.
		word_width = 0;
		while ((c = pgm_read_byte(s++))) {
			// Subtract 32 as our first character is number 32.
			c -= 32;
			// Calculate the offset to the character data in the font.
			data_offset = pgm_read_byte(ui_font + c * 2) + (pgm_read_byte(ui_font + c * 2 + 1) << 8);
			// Subtract the width of the character.
			word_width += pgm_read_byte(ui_font + data_offset);
			// Break at spaces.
			if (c == (' '-32)) break;
		}
		// Restore the start of the current word.
		s = t;
		
		// Do we need to wrap?
		if (word_width > width_remaining) {
			// Wrap.
			*x = original_x;
			*y += line_height;
			width_remaining = width - word_width;
		} else {
			// No need to wrap.
			width_remaining -= word_width;
		}
		
		// Draw the whole word.
		while ((c = pgm_read_byte(s++))) {
			ui_put_char(x, y, c);
			// Break at spaces.
			if (c == ' ') break;
		}
		
	} while (c);
	
}